/*
 *robotfindskitten: A Zen simulation
 *
 *Copyright (C) 1997,2000 Leonard Richardson 
 *                        leonardr@segfault.org
 *                        http://www.crummy.com/devel/
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or EXISTANCE OF KITTEN.  See the GNU General
 *   Public License for more details.
 *
 *   http://www.gnu.org/copyleft/gpl.html
 *
 *   cc65 port by Jeremy Penner (jeremypenner@shaw.ca)
 *   Successfully tested with C64 and Apple II.
 */

#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef __C64__
static char *ver = "C64!"; 
#elif defined (__APPLE2__)
static char *ver = "Woz Edition!";
#else
static char *ver = "Generic 6502!";
#endif

typedef enum { FALSE = 0, TRUE } bool; 

/*The messages go in a separate file because they are collectively
  huge, and you might want to modify them. It would be nice to load
  the messages from a text file at run time.*/
#include "messages.h"

/*Constants for our internal representation of the screen.*/
#define EMPTY -1
#define ROBOT 0
#define KITTEN 1

/*Screen dimensions.*/
#define X_MIN 0
#ifdef __APPLE2__
#define X_MAX 38
#else
#define X_MAX 39
#endif
#define Y_MIN 6
#define Y_MAX 23

/*Macros for generating numbers in different ranges*/
#define randx() (rand() % X_MAX+1)
#define randy() (rand() % (Y_MAX-Y_MIN+1))+Y_MIN /*I'm feeling randy()!*/
#define randchar() (rand() % (126-'!'+1)+'!')

/* We can't let the colour of anything be 6, the colour of the background! */
int randcolor(void)
{
  int col;

  while((col = rand() % 15 + 1) == 6);
  return (col);
}

/*Row constants for the animation*/
#define ADV_ROW 2
#define ANIMATION_MEET 25

/*This struct contains all the information we need to display an object
  on the screen*/
typedef struct
{
  int x;
  int y;
  int color;
  bool bold;
  char character;
} screen_object;

/*
 *Function definitions
 */

/*Initialization and setup functions*/
void initialize_arrays();
void initialize_robot();
void initialize_kitten();
void initialize_bogus();
void initialize_screen();
void instructions();

/*Game functions*/
void play_game();
void process_input(int);

/*Helper functions*/
int validchar(char);

void play_animation(int);

/*Global variables. Bite me, it's fun.*/
screen_object robot;
screen_object kitten;

int num_bogus;
screen_object bogus[MESSAGES];
int bogus_messages[MESSAGES];
int used_messages[MESSAGES];

/* This array contains our internal representation of the screen. The
 array is bigger than it needs to be, as we don't need to keep track
 of the first few rows of the screen. But that requires making an
 offset function and using that everywhere. So not right now. */
int screen[X_MAX+1][Y_MAX+1];

#include "draw6502.h"

/******************************************************************************
 *
 * Begin meaty routines that do the dirty work.
 *
 *****************************************************************************/

/*
 *play_game waits in a loop getting input and sending it to process_input
 */
void play_game()
{
  int old_x = robot.x;
  int old_y = robot.y;
  int input;

  input = cgetc();
  while ((input != 'q') && (input != 'Q'))
    {
      process_input(input);
      
      /*Redraw robot, where applicable. We're your station, robot.*/
      if (!(old_x == robot.x && old_y == robot.y))
	{
	  /*Get rid of the old robot*/
	  cputcxy(old_x,old_y, ' ');
	  screen[old_x][old_y] = EMPTY;
	  
	  /*Meet the new robot, same as the old robot.*/
	  draw(&robot);
	  screen[robot.x][robot.y] = ROBOT;
	  
	  old_x = robot.x;
	  old_y = robot.y;
	}
      input = cgetc();
    } 
  message("Bye!");
  gotoxy(0,22);
}

/*
 *Given the keyboard input, process_input interprets it in terms of moving,
 *touching objects, etc.
 */
void process_input(int input)
{
  int check_x = robot.x;
  int check_y = robot.y;

  switch (input)
    {
    case 'i':
    case 'I':
      check_y--;
      break;
    case 'u':
    case 'U':
      check_x--;
      check_y--;
      break;
    case 'o':
    case 'O':
      check_x++;
      check_y--;
      break;
    case ',':
      check_y++;
      break;
    case 'm':
    case 'M':
      check_x--;
      check_y++;
      break;
    case '.':
      check_x++;
      check_y++;
      break;
    case 'j':
    case 'J':
      check_x--;
      break;
    case 'l':
    case 'L':
      check_x++;
      break;
    case 0:
      break;
    default: /*Bad command or filename.*/
      message("Invalid input: Use direction keys or Esc.");
      return;
    }
  
  /*Check for going off the edge of the screen.*/
  if (check_y < Y_MIN || check_y > Y_MAX || check_x < X_MIN || check_x > X_MAX)
    {
      return; /*Do nothing.*/
    }
  
  /*Check for collision*/
  if (screen[check_x][check_y] != EMPTY)
    {
      switch (screen[check_x][check_y])
	{
	case ROBOT:
				/*We didn't move, or we're stuck in a
				  time warp or something.*/
	  break;
	case KITTEN: /*Found it!*/
          gotoxy(0,1);
          message(""); // clear the message area
	  play_animation(input);
	  break;
	default: /*We hit a bogus object; print its message.*/
	  message(messages[bogus_messages[screen[check_x][check_y]-2]]);
	  break;
	}
      return;
    }

  /*Otherwise, move the robot.*/
  robot.x = check_x;
  robot.y = check_y;
}

/******************************************************************************
 *
 * Begin helper routines
 *
 *****************************************************************************/

int validchar(char a)
{
  switch(a)
    {
    case '#':
    case ' ':   
    case 127:
      return 0;
    }
  return 1;
}

void play_animation(int input)
{
  int counter;
  int i;
  bool robotOnLeft;

  input = tolower(input);

  robotOnLeft = (input == 'l');
  robotOnLeft = robotOnLeft || (input == 'o');
  robotOnLeft = robotOnLeft || (input == '.');
  robotOnLeft = robotOnLeft || (input == ',');

  /*The grand cinema scene.*/
  for (counter = 4; counter >0; counter--)
    {
      /*Move the object on the right.*/
      cputcxy(ANIMATION_MEET+counter+1,ADV_ROW,' ');
      gotoxy(ANIMATION_MEET+counter, ADV_ROW);
      if (robotOnLeft)
	draw_in_place(&kitten);
      else
	draw_in_place(&robot);
      
      /*Move the object on the left.*/
      cputcxy(ANIMATION_MEET-counter,ADV_ROW,' ');
      gotoxy(ANIMATION_MEET-counter+1, ADV_ROW);
      if (robotOnLeft)
	draw_in_place(&robot);
      else
	draw_in_place(&kitten);

      // Busy loop(tm)!  For all your waiting needs!
      for(i = 0; i < 1000; i ++);
    }

  gotoxy(0,1);
  puts("You found kitten! Way to go, robot!");
  gotoxy(0,22);
  exit(0);
}

/******************************************************************************
 *
 * Begin initialization routines (called before play begins).
 *
 *****************************************************************************/

void instructions()
{
  char dummy;

  clrscr();
  gotoxy(0,0);
  printf("robotfindskitten %s\n",ver);
  printf("By the illustrious Leonard Richardson\n(C) 1997, 2000\n");
  printf("Ported to 6502 by Jeremy Penner\n<jeremypenner@shaw.ca>\n");
  printf("Written originally for the Nerth Pork\nrobotfindskitten contest\n\n");
  printf("In this game, you are robot (");
  draw_in_place(&robot);
  printf(").\nYour job is to find kitten. This task\n"); 
  printf("is complicated by the existance of\nvarious things which are not kitten.\n");  
  printf("Robot must touch items to determine if\nthey are kitten or not. The game\n");
  printf("ends when robotfindskitten.\nAlternatively, you may end the game by\n");
  printf("hitting the Q key.\n");
  printf("Robot is controlled with the 'I', 'J',\n'L', and ',' keys, as well\n");
  printf("as the standard 'U', 'M', 'O', and '.'\nfor diagonal movement.\n\n");
  printf("Press any key to start.\n");

  dummy = cgetc();
  clrscr();
}

void initialize_arrays()
{
  int counter, counter2;

  /*Initialize the empty object.*/
  screen_object empty;
  empty.x = -1;
  empty.y = -1;
  empty.color = 0;
  empty.bold = FALSE;
  empty.character = ' ';
  
  for (counter = 0; counter <= X_MAX; counter++)
    {
      for (counter2 = 0; counter2 <= Y_MAX; counter2++)
	{
	  screen[counter][counter2] = EMPTY;
	}
    }
  
  /*Initialize the other arrays.*/
  for (counter = 0; counter < MESSAGES; counter++)
    {
      used_messages[counter] = 0;
      bogus_messages[counter] = 0;
      bogus[counter] = empty;
    }
}

/*initialize_robot initializes robot.*/
void initialize_robot()
{
  /*Assign a position to the player.*/
  robot.x = randx();
  robot.y = randy();

  robot.character = '#';
  robot.color = 0;
  robot.bold = FALSE;
  screen[robot.x][robot.y] = ROBOT;
}

/*initialize kitten, well, initializes kitten.*/
void initialize_kitten()
{
  /*Assign the kitten a unique position.*/
  do
    {
      kitten.x = randx();
      kitten.y = randy();
    } while (screen[kitten.x][kitten.y] != EMPTY);
  
  /*Assign the kitten a character and a color.*/
  do {
    kitten.character = randchar();
  } while (!(validchar(kitten.character))); 
  screen[kitten.x][kitten.y] = KITTEN;

  kitten.color = randcolor();
}

/*initialize_bogus initializes all non-kitten objects to be used in this run.*/
void initialize_bogus()
{
  int counter, index;
  for (counter = 0; counter < num_bogus; counter++)
    {
      /*Give it a color.*/
      bogus[counter].color = randcolor();
      
      /*Give it a character.*/
      do {
	bogus[counter].character = randchar();
      } while (!(validchar(bogus[counter].character))); 
      
      /*Give it a position.*/
      do
	{
	  bogus[counter].x = randx();
	  bogus[counter].y = randy();
	} while (screen[bogus[counter].x][bogus[counter].y] != EMPTY);

      screen[bogus[counter].x][bogus[counter].y] = counter+2;
      
      /*Find a message for this object.*/
      do {
	index = rand() % MESSAGES;
      } while (used_messages[index] != 0);
      
      bogus_messages[counter] = index;
      used_messages[index] = 1;
    }

}

/*initialize_screen paints the screen.*/
void initialize_screen()
{
  int counter;

  /*
   *Print the status portion of the screen.
   */
  gotoxy(0,0);
  printf("robotfindskitten %s\n\n",ver);  
  
  gotoxy(0,Y_MIN - 1);
  /*Draw a line across the screen.*/
  for (counter = X_MIN; counter <= X_MAX; counter++)
    {
      printf("-");
    }

  /*
   *Draw all the objects on the playing field.
   */
  for (counter = 0; counter < num_bogus; counter++)
    {
      draw(&bogus[counter]);
    }

  draw(&kitten);
  draw(&robot);
}

int main(void)
{
  /*
   *Do general start-of-program stuff.
   */

  /*Get a number of non-kitten objects.*/
  num_bogus = 15;

  /*Initialize the random number generator*/
#ifdef __APPLE2__
  srand(*((int *)0x4E));
#else
  //srand(clock());
  srand(1);
#endif

  initialize_arrays();

  /*
   *Now we initialize the various game objects.
   */
  initialize_robot();
  initialize_kitten();
  initialize_bogus();

  instructions();  

  initialize_screen();
  play_game();
}
